@@ -5,7 +5,7 @@
     use Symbol;
 # ----------------------------------------------------------------
     use vars qw( $VERSION );
-    $VERSION = "0.07";
+    $VERSION = "0.08";
 # ----------------------------------------------------------------
     my $XML_ENCODING = "UTF-8";
     my $INTERNAL_ENCODING = "UTF-8";
@@ -148,14 +148,24 @@ The result dumped is following:
 The special node name of "#text" is used because this elements
 has attribute(s) in addition to the text node.
 
-=head1 METHODS
+=head1 CONSTRUCTOR AND OPTIONS
 
-=head2 $tpp = XML::TreePP->new( %attributes )
+=head2 $tpp = XML::TreePP->new();
 
-This method constructs a new "XML::TreePP" object.
-Optional attributes are following:
+This constructor method returns a new "XML::TreePP" object.
 
-=head2 output_encoding => "UTF-8",
+=head2 $tpp = XML::TreePP->new( %options );
+
+Its first argument is a hash variable to set one or more options 
+like following:
+
+=head2 $tpp->set( option_name => $option_value );
+
+This method sets a option value for "option_name".
+If $option_value is not defined, its option is deleted.
+Options below are available:
+
+=head2 $tpp->set( output_encoding => "UTF-8" );
 
 You can define a encoding of xml file generated by write/writefile
 methods. On Perl 5.8.x and later, you can select it from every
@@ -163,24 +173,36 @@ encodings supported by Encode.pm. On Perl 5.6.x or before with
 Jcode.pm, you can use "Shift_JIS", "EUC-JP", "ISO-2022-JP" and
 "UTF-8". The default value is "UTF-8".
 
-=head2 force_array => [ "rdf:li", "item", "-xmlns" ],
+=head2 $tpp->set( force_array => [ "rdf:li", "item", "-xmlns" ] );
 
 This option allows you to specify a list of element names which
 should always be forced into an array representation
 The default value is null, it means that context of the elements
 will determine to make array or to keep it scalar.
 
-=head2 first_out => [ "link", "title", "-type" ],
+=head2 $tpp->set( first_out => [ "link", "title", "-type" ] );
 
 This option allows you to specify a list of element/attribute
 names which should always appears at first on output XML code.
 The default value is null, it means alphabetical order is used.
 
-=head2 last_out => [ "items", "item", "entry" ],
+=head2 $tpp->set( last_out => [ "items", "item", "entry" ] );
 
 This option allows you to specify a list of element/attribute
 names which should always appears at last on output XML code.
 
+=head2 $tpp->set( cdata_scalar_ref => 1 );
+
+This option allows you to convert a cdata section into a reference
+for scalar on parsing XML source. If this option is false, per 
+default, cdata section is converted into a scalar.
+
+=head2 $tpp->get( "option_name" );
+
+This method returns a current option value for "option_name".
+
+=head1 METHODS
+
 =head2 $tree = $tpp->parse( $source );
 
 This method reads XML source and returns a hash tree converted.
@@ -213,8 +235,7 @@ $tree is a referecen to a hash tree.
 
 =head1 AUTHOR
 
-Yusuke Kawasaki, E<lt>u-suke [at] kawa.netE<gt>
-http://www.kawa.net/works/perl/treepp/treepp-e.html
+Yusuke Kawasaki, http://www.kawa.net/
 
 =head1 COPYRIGHT AND LICENSE
 
@@ -231,18 +252,36 @@ sub new {
     $self;
 }
 # ----------------------------------------------------------------
+sub set {
+    my $self = shift;
+    my $key = shift;
+    my $val = shift;
+    if ( defined $val ) {
+        $self->{$key} = $val;
+    } else {
+        delete $self->{$key};
+    }
+}
+# ----------------------------------------------------------------
+sub get {
+    my $self = shift;
+    my $key = shift;
+    $self->{$key} if exists $self->{$key};
+}
+# ----------------------------------------------------------------
 sub writefile {
     my $self = shift;
     my $file = shift;
-    my $tree = shift;
+    my $tree = shift or Carp::croak "Invalid tree";
     my $encode = shift;
+    Carp::croak "Invalid filename" unless defined $file;
     my $text = $self->write( $tree, $encode );
     &write_raw_xml( $file, $text );
 }
 # ----------------------------------------------------------------
 sub write {
     my $self = shift;
-    my $tree = shift;
+    my $tree = shift or Carp::croak "Invalid tree";
     my $from = $self->{internal_encoding} || $INTERNAL_ENCODING;
     my $to = shift || $self->{output_encoding} || $XML_ENCODING;
     my $decl = '<?xml version="1.0" encoding="'.$to.'" ?>';
@@ -282,8 +321,8 @@ sub parsehttp {
 # ----------------------------------------------------------------
 sub parsehttp_lwp {
     my $self = shift;
-    my $method = shift;
-    my $url = shift;
+    my $method = shift or Carp::croak "Invalid HTTP method";
+    my $url = shift or Carp::croak "Invalid URL";
     my $body = shift;
 
     my $ua = LWP::UserAgent->new();
@@ -299,8 +338,8 @@ sub parsehttp_lwp {
 # ----------------------------------------------------------------
 sub parsehttp_lite {
     my $self = shift;
-    my $method = shift;
-    my $url = shift;
+    my $method = shift or Carp::croak "Invalid HTTP method";
+    my $url = shift or Carp::croak "Invalid URL";
     my $body = shift;
 
     my $http = HTTP::Lite->new();
@@ -314,6 +353,7 @@ sub parsehttp_lite {
 sub parsefile {
     my $self = shift;
     my $file = shift;
+    Carp::croak "Invalid filename" unless defined $file;
     my $text = &read_raw_xml( $file );
     $self->parse( \$text );
 }
@@ -321,6 +361,7 @@ sub parsefile {
 sub parse {
     my $self = shift;
     my $textref = ref $_[0] ? $_[0] : \$_[0];
+    Carp::croak "Invalid XML source" if ( ref($textref) ne "SCALAR" );
     my $to = $self->{internal_encoding} || $INTERNAL_ENCODING;
     if ( $to ) {
         my $from = &xml_decl_encoding( $textref );
@@ -362,9 +403,8 @@ sub xml_to_flat {
             $typeDocT, $contDocT,
             $typeCmnt, $contCmnt, $typeElem, $contElem, $follow ) =
             ( $1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13 );
-        if ( $ahead ne "" ) {
-            ## warn "Invalid header ignored: [$ahead]\n";
-            print "Invalid header ignored: [$ahead] before <$match>\n";
+        if ( defined $ahead && $ahead =~ /\S/ ) {
+            Carp::carp "Invalid string: [$ahead] before <$match>";
         }
 
         if ( $typeElem ) {                          ## Element
@@ -384,11 +424,15 @@ sub xml_to_flat {
                     my $val = &xml_unescape( $2 ? $3 : $4 );
                     $attr->{"-".$key} = $val;
                 }
-                $node->{attributes} = $attr;
+                $node->{attributes} = $attr if scalar keys %$attr;
             }
             push( @$flat, $node );
         } elsif ( $typeCDATA ) {                    ## CDATASection
-            push( @$flat, $contCDATA );
+            if ( exists $self->{cdata_scalar_ref} && $self->{cdata_scalar_ref} ) {
+                push( @$flat, \$contCDATA );        # as reference for scalar
+            } else {
+                push( @$flat, $contCDATA );         # as scalar like text node
+            }
         } elsif ( $typeCmnt ) {                     ## Comment
             # ignore
         } elsif ( $typeDocT ) {                     ## DocumentType
@@ -396,9 +440,9 @@ sub xml_to_flat {
         } elsif ( $typePI ) {                       ## ProcessingInstruction
             # ignore
         } else {
-            warn "Invalid Tag: $match\n";
+            Carp::carp "Invalid Tag: <$match>";
         }
-        if ( $follow =~ /\S/ ) {
+        if ( $follow =~ /\S/ ) {                    ## text node
             $follow =~ s/^([ \t]*[\r\n])+//s;
             $follow =~ s/([\r\n][ \t]*)+$//s;
             my $val = &xml_unescape( $follow );
@@ -417,8 +461,9 @@ sub flat_to_tree {
 
     while( scalar @$source ) {
         my $node = shift @$source;
-        if ( ! ref $node ) {
-            push( @$text, $node );
+        # print Data::Dumper::Dumper( $node );
+        if ( ! ref $node || UNIVERSAL::isa( $node, "SCALAR" )) {
+            push( @$text, $node );                  ## cdata or text node
             next;
         }
         my $name = $node->{tagName};
@@ -427,7 +472,7 @@ sub flat_to_tree {
             Carp::croak "Invalid tag sequence: <$parent></$name>";
         }
         my $elem = $node->{attributes};
-        if ( $node->{startTag} ) {
+        if ( $node->{startTag} ) {                  ## recursive call
             my $child = $self->flat_to_tree( $source, $name );
             if ( ref $elem && scalar keys %$elem ) {
                 if ( ref $child ) {
@@ -444,7 +489,7 @@ sub flat_to_tree {
                 $elem = $child;
             }
         }
-        next unless defined $elem;
+#       next unless defined $elem;
         $tree->{$name} ||= [];
         push( @{$tree->{$name}}, $elem );
     }
@@ -486,6 +531,9 @@ sub hash_to_xml {
             } elsif ( UNIVERSAL::isa( $val, "ARRAY" ) ) {
                 my $child = $self->array_to_xml( $key, $val );
                 push( @$out, $child );
+            } elsif ( UNIVERSAL::isa( $val, "SCALAR" ) ) {
+                my $child = $self->scalaref_to_cdata( $key, $val );
+                push( @$out, $child );
             } elsif ( ref $val ) {
                 my $child = $self->hash_to_xml( $key, $val );
                 push( @$out, $child );
@@ -501,7 +549,7 @@ sub hash_to_xml {
         }
     }
     my $jattr = join( "", @$attr );
-#   s/^(\s*<)/  $1/mg foreach @$out;
+#   s/^(\s*<)/  $1/mg foreach @$out;            # indent
     my $text = join( "", @$out );
     if ( defined $name ) {
         if ( scalar @$out ) {
@@ -528,6 +576,9 @@ sub array_to_xml {
         } elsif ( UNIVERSAL::isa( $val, "ARRAY" ) ) {
             my $child = $self->array_to_xml( $name, $val );
             push( @$out, $child );
+        } elsif ( UNIVERSAL::isa( $val, "SCALAR" ) ) {
+            my $child = $self->scalaref_to_cdata( $name, $val );
+            push( @$out, $child );
         } elsif ( ref $val ) {
             my $child = $self->hash_to_xml( $name, $val );
             push( @$out, $child );
@@ -536,11 +587,20 @@ sub array_to_xml {
             push( @$out, $child );
         }
     }
-#   s/^(\s*<)/  $1/mg foreach @$out;
+#   s/^(\s*<)/  $1/mg foreach @$out;            # indent
     my $text = join( "", @$out );
     $text;
 }
 # ----------------------------------------------------------------
+sub scalaref_to_cdata {
+    my $self = shift;
+    my $name = shift;
+    my $ref = shift;
+    my $text = '<![CDATA['.$$ref.']]>';
+    $text = "<$name>$text</$name>\n" if ( $name ne "#text" );
+    $text;
+}
+# ----------------------------------------------------------------
 sub scalar_to_xml {
     my $self = shift;
     my $name = shift;
